//==================================
// PEDUMP - Matt Pietrek 1997
// FILE: EXEDUMP.C
//==================================

#include "Common.h"
#include "COFFSymbolTable.h"
#include "ExtrnVar.h"

#if defined(__MINGW32__) && !defined(__MINGW64_VERSION_MAJOR)
 /*
  * '__MINGW32__' is defined by all these packages:
  *    the old-school MinGW      [1].
  *    the more modern MinGW-w64 [2] and TDM-gcc [3].
  *
  * [1] http://mingw.org/
  * [2] http://mingw-w64.sourceforge.net/
  * [3] http://tdm-gcc.tdragon.net/
  *
  * But only [2] + [3] defines '__MINGW64_VERSION_MAJOR'. Hence:
  */
  #define OLD_SCHOOL_MINGW
#endif

#ifdef __MINGW32__
  #define IMAGE_LOAD_CONFIG_DIRECTORY32 IMAGE_LOAD_CONFIG_DIRECTORY
#endif


// Names of the data directory elements that are defined
static const char *ImageDirectoryNames[] = {
    "EXPORT", "IMPORT", "RESOURCE", "EXCEPTION", "SECURITY", "BASERELOC",
    "DEBUG", "COPYRIGHT", "GLOBALPTR", "TLS", "LOAD_CONFIG",
    "BOUND_IMPORT", "IAT",  // These two entries added for NT 3.51
    "DELAY_IMPORT" };       // This entry added in NT 5

// Marked as obsolete in MSDN CD 9
// Bitfield values and names for the DllCharacteritics flags
static WORD_FLAG_DESCRIPTIONS DllCharacteristics[] = {
                    { IMAGE_DLLCHARACTERISTICS_WDM_DRIVER, "WDM_DRIVER" }
                  };

#define NUMBER_IMAGE_DIRECTORY_ENTRYS  DIM (ImageDirectoryNames)
#define NUMBER_DLL_CHARACTERISTICS     DIM (DllCharacteristics)

static void DisplayDataDirectoryEntry (const char *pszName, const IMAGE_DATA_DIRECTORY *dataDirEntry)
{
  printf ("  %-16s rva: %08lX  size: %08lX\n", pszName, dataDirEntry->VirtualAddress, dataDirEntry->Size);
}

//
// Given a section name, look it up in the section table and return a
// pointer to its IMAGE_SECTION_HEADER
//
template <class T>
static const IMAGE_SECTION_HEADER *GetSectionHeader (const char *name, const T* pNTHeader)
{
  const IMAGE_SECTION_HEADER *section = IMAGE_FIRST_SECTION (pNTHeader);
  UINT  i;

  for (i = 0; i < pNTHeader->FileHeader.NumberOfSections; i++, section++)
  {
    const char *sec_name = (const char*)section->Name;

    if (!strncmp(sec_name,name,IMAGE_SIZEOF_SHORT_NAME))
        return (section);
  }
  return 0;
}

template <class T>
static void DumpExeDebugDirectory (const BYTE *base, const T *pNTHeader)
{
  const IMAGE_DEBUG_DIRECTORY *debugDir;
  const IMAGE_SECTION_HEADER  *header;
  DWORD size, va_debug_dir = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_DEBUG);

  if (va_debug_dir == 0)
     return;

  // If we found a .debug section, and the debug directory is at the
  // beginning of this section, it looks like a Borland file
  header = GetSectionHeader (".debug", pNTHeader);
  if (header && header->VirtualAddress == va_debug_dir)
  {
    debugDir = (const IMAGE_DEBUG_DIRECTORY*) (header->PointerToRawData + base);
    size = GetImgDirEntrySize (pNTHeader, IMAGE_DIRECTORY_ENTRY_DEBUG) *
                               sizeof(IMAGE_DEBUG_DIRECTORY);
  }
  else    // Look for the debug directory
  {
    header = GetEnclosingSectionHeader (va_debug_dir, pNTHeader);
    if (!header)
       return;

    size = GetImgDirEntrySize (pNTHeader, IMAGE_DIRECTORY_ENTRY_DEBUG);

    debugDir = MakePtr (const IMAGE_DEBUG_DIRECTORY *, base,
                        header->PointerToRawData + va_debug_dir -
                        header->VirtualAddress);
  }
  DumpDebugDirectory (debugDir, size, base);
}


template <class T, class U, class V>
static void DumpImportsOfOneModule (const T                       *pINT,
                                    const U                       *pIAT,
                                    const V                       *pNTHeader,
                                    const IMAGE_IMPORT_DESCRIPTOR *pImportDesc,
                                    const BYTE                    *imageBase)
{
  const IMAGE_IMPORT_BY_NAME *pOrdinalName;
  bool  bIs64Bit = (pNTHeader->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC);

  while (1) // Loop forever (or until we break out)
  {
    if (pINT->u1.AddressOfData == 0)
       break;

    ULONGLONG ordinal = -1;

    if (bIs64Bit)
    {
      if (IMAGE_SNAP_BY_ORDINAL64(pINT->u1.Ordinal))
         ordinal = IMAGE_ORDINAL64 (pINT->u1.Ordinal);
    }
    else
    {
      if (IMAGE_SNAP_BY_ORDINAL32(pINT->u1.Ordinal))
         ordinal = IMAGE_ORDINAL32 (pINT->u1.Ordinal);
    }

    if (ordinal != (ULONGLONG)-1)
    {
      printf ("  %4I64u", ordinal);
    }
    else
    {
      // pINT->u1.AddressOfData is theoretically 32 or 64 bits, but in the on-disk representation,
      // we'll assume it's an RVA.  As such, we'll cast it to a DWORD.
      pOrdinalName = (const IMAGE_IMPORT_BY_NAME*) GetPtrFromRVA (
                       static_cast<DWORD>(pINT->u1.AddressOfData), pNTHeader, imageBase);

        printf ("  %4u  %s",
                pOrdinalName->Hint,
                undecorate_name((const char*)&pOrdinalName->Name[0]));
    }

    // If it looks like the image has been bound, append the bound address
    if (pImportDesc->TimeDateStamp)
       printf (" (Bound to: %08I64X)", (ULONGLONG)pIAT->u1.Function);
    printf ("\n");

    pINT++;         // Advance to next thunk
    pIAT++;         // advance to next thunk
  }
}


//
// Dump the imports table (the .idata section) of a PE file
//
template <class T>
static void DumpImportsSection (const BYTE *imageBase,
                                const T    *pNTHeader)
{
    const IMAGE_IMPORT_DESCRIPTOR *pImportDesc;
    DWORD importsStartRVA;

    // Look up where the imports section is (normally in the .idata section)
    // but not necessarily so.  Therefore, grab the RVA from the data dir.
    importsStartRVA = GetImgDirEntryRVA (pNTHeader,IMAGE_DIRECTORY_ENTRY_IMPORT);
    if (!importsStartRVA)
       return;

    pImportDesc = (const IMAGE_IMPORT_DESCRIPTOR*) GetPtrFromRVA (importsStartRVA, pNTHeader, imageBase);
    if (!pImportDesc)
       return;

    bool bIs64Bit = (pNTHeader->FileHeader.SizeOfOptionalHeader == IMAGE_SIZEOF_NT_OPTIONAL64_HEADER);

    printf ("Imports Table:\n");

    while (1)
    {
        // See if we've reached an empty IMAGE_IMPORT_DESCRIPTOR
        if (pImportDesc->TimeDateStamp == 0 && pImportDesc->Name == 0)
           break;

        const char *imp_name = (const char*)GetPtrFromRVA(pImportDesc->Name, pNTHeader, imageBase);
        printf ("  %s\n", imp_name);

        printf ("  Import Lookup Table RVA:  %08lX\n", pImportDesc->Characteristics);
        printf ("  TimeDateStamp:            %08lX", pImportDesc->TimeDateStamp);

        if (pImportDesc->TimeDateStamp && pImportDesc->TimeDateStamp != 0xFFFFFFFF)
        {
           time_t timeStamp = pImportDesc->TimeDateStamp;
           printf (" -> %s", ctime(&timeStamp));
        }
        else
           printf ("\n");

        printf ("  ForwarderChain:           %08lX\n", pImportDesc->ForwarderChain);
        printf ("  DLL Name RVA:             %08lX\n", pImportDesc->Name);
        printf ("  Import Address Table RVA: %08lX\n", pImportDesc->FirstThunk);

        DWORD rvaINT = pImportDesc->OriginalFirstThunk;
        DWORD rvaIAT = pImportDesc->FirstThunk;

        if (rvaINT == 0)   // No Characteristics field?
        {
            // Yes! Gotta have a non-zero FirstThunk field then.
            rvaINT = rvaIAT;

            if (rvaINT == 0)   // No FirstThunk field?  Ooops!!!
               return;
        }

        // Adjust the pointer to point where the tables are in the
        // mem mapped file.
        const IMAGE_THUNK_DATA *pINT = (const IMAGE_THUNK_DATA*) GetPtrFromRVA (rvaINT, pNTHeader, imageBase);
        if (!pINT)
           return;

        const IMAGE_THUNK_DATA *pIAT = (const IMAGE_THUNK_DATA*) GetPtrFromRVA (rvaIAT, pNTHeader, imageBase);

        printf ("  Ordn  Name\n");

        bIs64Bit ? DumpImportsOfOneModule ((const IMAGE_THUNK_DATA64*)pINT,
                                           (const IMAGE_THUNK_DATA64*)pIAT,
                                           pNTHeader, pImportDesc, imageBase)
                 : DumpImportsOfOneModule ((const IMAGE_THUNK_DATA32*)pINT,
                                           (const IMAGE_THUNK_DATA32*)pIAT,
                                           pNTHeader, pImportDesc, imageBase);

        pImportDesc++;   // advance to next IMAGE_IMPORT_DESCRIPTOR
        printf ("\n");
    }
    printf ("\n");
}


template <class T>
static void DumpDelayedImportsSection (const BYTE *imageBase,
                                       const T    *pNTHeader,
                                       bool        bIs64Bit)
{
    const struct ImgDelayDescr *pDelayDesc;
    DWORD                       delayImportStartRVA;

    // Look up where the delay imports section is (normally in the .didat
    /// section) but not necessarily so.
    delayImportStartRVA = GetImgDirEntryRVA(pNTHeader, IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT);
    if (!delayImportStartRVA)
        return;

    // This code is more complicated than it needs to be, thanks to Microsoft.  When the
    // ImgDelayDescr was originally created for Win32, portability to Win64 wasn't
    // considered.  As such, MS used pointers, rather than RVAs in the data structures.
    // Finally, MS issued a new DELAYIMP.H, which issued a flag indicating whether the
    // field values are RVAs or VAs.  Unfortunately, Microsoft has been rather slow to
    // get this header file out into general distibution.  Currently, you can get it as
    // part of the Win64 headers, or as part of VC7.  In the meanwhile, we'll use some
    // preprocessor trickery so that we can use the new field names, while still compiling
    // with the original DELAYIMP.H.

    pDelayDesc = (const struct ImgDelayDescr*) GetPtrFromRVA (delayImportStartRVA,
                                                              pNTHeader, imageBase);
    if (!pDelayDesc)
        return;

    printf ("Delay Imports Table:\n");

    while (pDelayDesc->rvaDLLName)
    {
        const char *pszDLLName;

        // from more recent DELAYIMP.H:
        // enum DLAttr {                   // Delay Load Attributes
        //    dlattrRva = 0x1,             // RVAs are used instead of pointers
        //    };
        bool bUsingRVA = pDelayDesc->grAttrs & 1;

        DWORD dllNameRVA = (DWORD)pDelayDesc->rvaDLLName;
        PVOID dllNameVA = (PBYTE)0 + (DWORD)pDelayDesc->rvaDLLName;

        pszDLLName = bUsingRVA
            ? (const char*) GetPtrFromRVA (dllNameRVA, pNTHeader, imageBase)
            : (const char*) GetPtrFromVA (dllNameVA, pNTHeader, imageBase);

        if (!pszDLLName)
        {
           printf ("  No .DLL for ordinals!\n");
           break;
        }

        printf ("  %s\n", pszDLLName);
        printf ("  Attributes:                 %08lX", pDelayDesc->grAttrs);
        if (pDelayDesc->grAttrs & 1)
            printf ("  dlattrRva");
        printf ("\n");
        printf ("  Name R(VA):                 %08lX\n", pDelayDesc->rvaDLLName);
        printf ("  HMODULE R(VA):              %08lX\n", pDelayDesc->rvaHmod);
        printf ("  Import Address Table R(VA): %08lX\n", pDelayDesc->rvaIAT);
        printf ("  Import Names Table R(VA):   %08lX\n", pDelayDesc->rvaINT);
        printf ("  Bound IAT R(VA):            %08lX\n", pDelayDesc->rvaBoundIAT);
        printf ("  Unload IAT R(VA):           %08lX\n", pDelayDesc->rvaUnloadIAT);
        printf ("  TimeDateStamp:              %08lX", pDelayDesc->dwTimeStamp);

        if (pDelayDesc->dwTimeStamp)
        {
          time_t timeStamp = pDelayDesc->dwTimeStamp;
          printf (" -> %s", ctime(&timeStamp));
        }
        else
          printf ("\n");

        //
        // Display the Import Names Table.
        const void *pvThunk;
        const void *thunkVA = (PBYTE)0 + (DWORD)pDelayDesc->rvaINT;

        if (pDelayDesc->grAttrs & 1)  // see above comment on dlatttrRva being == 1
             pvThunk = GetPtrFromRVA ((DWORD)pDelayDesc->rvaINT, pNTHeader, imageBase);
        else pvThunk = GetPtrFromVA (thunkVA, pNTHeader, imageBase);

        printf ("  Ordn  Name\n");

        bIs64Bit ? DumpDelayedImportsImportNames (imageBase, pNTHeader,
                                                  (const IMAGE_THUNK_DATA64*)pvThunk,
                                                  bUsingRVA)
                 : DumpDelayedImportsImportNames (imageBase, pNTHeader,
                                                  (const IMAGE_THUNK_DATA32*)pvThunk,
                                                  bUsingRVA);

        pDelayDesc++;   // Pointer math.  Advance to next delay import desc.
    }
    printf ("\n");
}

//
// Dump the delayed imports table of a PE file
//
template <class T, class U>
static void DumpDelayedImportsImportNames (const BYTE *imageBase,
                                           const T    *pNTHeader,
                                           const U    *thunk,
                                           bool        bUsingRVA)
{
    while (1) // Loop forever (or until we break out)
    {
        if (thunk->u1.AddressOfData == 0)
           break;

        ULONGLONG ordinalMask;

        if (sizeof(thunk->u1.Ordinal) == sizeof(ULONGLONG))  // Which ordinal mask should we use?
             ordinalMask = IMAGE_ORDINAL_FLAG64;
        else ordinalMask = IMAGE_ORDINAL_FLAG32;

        if (thunk->u1.Ordinal & ordinalMask)
        {
            printf ("  %4u", (WORD)(thunk->u1.Ordinal & 0xFFFF));
        }
        else
        {
            const IMAGE_IMPORT_BY_NAME *pOrdinalName = (const IMAGE_IMPORT_BY_NAME*) ((BYTE*)0 + thunk->u1.AddressOfData);

            pOrdinalName = bUsingRVA
              ? (const IMAGE_IMPORT_BY_NAME*) GetPtrFromRVA ((DWORD)thunk->u1.AddressOfData, pNTHeader, imageBase)
              : (const IMAGE_IMPORT_BY_NAME*) GetPtrFromVA (pOrdinalName, pNTHeader, imageBase);

            printf ("  %4u  %s", pOrdinalName->Hint, pOrdinalName->Name);
        }
        printf ("\n");
        thunk++;            // Advance to next thunk
    }
    printf ("\n");
}


//
// Dump the exports table (usually the .edata section) of a PE file
//
template <class T>
static void DumpExportsSection (const BYTE *imageBase, const T *pNTHeader)
{
    const IMAGE_EXPORT_DIRECTORY *pExportDir;
    const IMAGE_SECTION_HEADER   *header;
    const char  *pszFilename;
    DWORD        i, exportsStartRVA, exportsEndRVA;
    const DWORD *pdwFunctions;
    const WORD  *pwOrdinals;
    const DWORD *pszFuncNames;

    exportsStartRVA = GetImgDirEntryRVA(pNTHeader,IMAGE_DIRECTORY_ENTRY_EXPORT);
    exportsEndRVA = exportsStartRVA +
                    GetImgDirEntrySize(pNTHeader, IMAGE_DIRECTORY_ENTRY_EXPORT);

    /* Get the IMAGE_SECTION_HEADER that contains the exports.  This is
     * usually the .edata section, but doesn't have to be.
     */
    header = GetEnclosingSectionHeader (exportsStartRVA, pNTHeader);
    if (!header)
       return;

    pExportDir  = (const IMAGE_EXPORT_DIRECTORY*) GetPtrFromRVA (exportsStartRVA, pNTHeader, imageBase);
    pszFilename = (const char *) GetPtrFromRVA (pExportDir->Name, pNTHeader, imageBase);

    printf ("exports table:\n\n");
    printf ("  Name:            %s\n", pszFilename);
    printf ("  Characteristics: %08lX\n", pExportDir->Characteristics);

    time_t timeStamp = *(DWORD*) &pExportDir->TimeDateStamp;
    const char *t_str = "<?>";

    if (timeStamp > 0 && timeStamp < 0xFFFFFFFF)
       t_str = ctime(&timeStamp);
    printf ("  TimeDateStamp:   %08lX -> %.24s\n", pExportDir->TimeDateStamp, t_str);
    printf ("  Version:         %u.%02u\n", pExportDir->MajorVersion, pExportDir->MinorVersion);

    printf ("  Ordinal base:    %08lX\n", pExportDir->Base);
    printf ("  # of functions:  %08lX\n", pExportDir->NumberOfFunctions);
    printf ("  # of Names:      %08lX\n", pExportDir->NumberOfNames);

    pdwFunctions = (const DWORD*) GetPtrFromRVA (pExportDir->AddressOfFunctions, pNTHeader, imageBase);
    pwOrdinals   = (const WORD*)  GetPtrFromRVA (pExportDir->AddressOfNameOrdinals, pNTHeader, imageBase);
    pszFuncNames = (const DWORD*) GetPtrFromRVA (pExportDir->AddressOfNames, pNTHeader, imageBase);

    printf ("\n  Entry Pt  Ordn  Name\n");

    for  (i = 0; i < pExportDir->NumberOfFunctions; i++, pdwFunctions++)
    {
        DWORD entryPointRVA = *pdwFunctions;

        if (entryPointRVA == 0)   // Skip over gaps in exported function
           continue;              // ordinals (the entrypoint is 0 for
                                  // these functions).

        printf ("  %08lX  %4lu", entryPointRVA, i + pExportDir->Base);

        // See if this function has an associated name exported for it.
        for (unsigned j = 0; j < pExportDir->NumberOfNames; j++)
            if (pwOrdinals[j] == i)
            {
               const char *name = (const char *) GetPtrFromRVA (pszFuncNames[j], pNTHeader, imageBase);
               printf ("  %s", undecorate_name(name));
            }

        // Is it a forwarder?  If so, the entry point RVA is inside the
        // .edata section, and is an RVA to the DllName.EntryPointName
        if (entryPointRVA >= exportsStartRVA && entryPointRVA <= exportsEndRVA)
           printf (" (forwarder -> %s)", (char*) GetPtrFromRVA(entryPointRVA, pNTHeader, imageBase));
        printf ("\n");
    }
    printf ("\n");
}


//
// Dump the TLS data
//
static void DumpTLSDirectory32 (const BYTE *imageBase,
                                const IMAGE_NT_HEADERS32 *pNTHeader)
{
    const IMAGE_TLS_DIRECTORY32 *pTLSDir;
    DWORD tlsRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_TLS);

    if (!tlsRVA)
      return;

    pTLSDir = (const IMAGE_TLS_DIRECTORY32*) GetPtrFromRVA (tlsRVA, pNTHeader, imageBase);

    printf ("TLS directory:\n");
    printf ("  StartAddressOfRawData: %08lX\n", pTLSDir->StartAddressOfRawData);
    printf ("  EndAddressOfRawData:   %08lX\n", pTLSDir->EndAddressOfRawData);
    printf ("  AddressOfIndex:        %08lX\n", pTLSDir->AddressOfIndex);
    printf ("  AddressOfCallBacks:    %08lX\n", pTLSDir->AddressOfCallBacks);
    printf ("  SizeOfZeroFill:        %08lX\n", pTLSDir->SizeOfZeroFill);
    printf ("  Characteristics:       %08lX\n", pTLSDir->Characteristics);
    printf ("\n");
}

static void DumpTLSDirectory64 (const BYTE *imageBase,
                                const IMAGE_NT_HEADERS64 *pNTHeader)
{
    const IMAGE_TLS_DIRECTORY64 *pTLSDir;
    DWORD tlsRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_TLS);

    if (!tlsRVA)
      return;

    pTLSDir = (const IMAGE_TLS_DIRECTORY64*) GetPtrFromRVA (tlsRVA, pNTHeader, imageBase);

    printf ("TLS directory:\n");
    printf ("  StartAddressOfRawData: %08I64X\n", pTLSDir->StartAddressOfRawData);
    printf ("  EndAddressOfRawData:   %08I64X\n", pTLSDir->EndAddressOfRawData);
    printf ("  AddressOfIndex:        %08I64X\n", pTLSDir->AddressOfIndex);
    printf ("  AddressOfCallBacks:    %08I64X\n", pTLSDir->AddressOfCallBacks);
    printf ("  SizeOfZeroFill:        %08lX\n", pTLSDir->SizeOfZeroFill);
    printf ("  Characteristics:       %08lX\n", pTLSDir->Characteristics);
    printf ("\n");
}


template <class T>
static void DumpRuntimeFunctions (const BYTE *imageBase, const T *pNTHeader)
{
    DWORD rtFnRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_EXCEPTION);
    DWORD i, cEntries;

    if (!rtFnRVA)
        return;

    cEntries = GetImgDirEntrySize (pNTHeader, IMAGE_DIRECTORY_ENTRY_EXCEPTION) /
                                   sizeof(IMAGE_RUNTIME_FUNCTION_ENTRY);
    if (cEntries == 0)
        return;

    const IMAGE_IA64_RUNTIME_FUNCTION_ENTRY *pRTFn = (const IMAGE_IA64_RUNTIME_FUNCTION_ENTRY*)
                                                     GetPtrFromRVA (rtFnRVA, pNTHeader, imageBase);
    if (!pRTFn)
        return;

    printf ("Runtime Function Table (Exception handling)\n");
    printf ("  Begin     End       Unwind  \n");
    printf ("  --------  --------  --------\n");

    for (i = 0; i < cEntries; i++, pRTFn++)
    {
      printf ("  %08lX  %08lX  %08lX",
              pRTFn->BeginAddress, pRTFn->EndAddress, pRTFn->UnwindInfoAddress);

      if (g_pCOFFSymbolTable)
      {
        COFFSymbol *pSymbol = g_pCOFFSymbolTable->GetNearestSymbolFromRVA (pRTFn->BeginAddress, TRUE);
        if (pSymbol)
           printf ("  %s", pSymbol->GetName());
        delete pSymbol;
      }
      printf ("\n");
    }
}


// The names of the available base relocations
static const char *SzRelocTypes[] = {
                  "ABSOLUTE","HIGH","LOW","HIGHLOW","HIGHADJ","MIPS_JMPADDR",
                  "SECTION","REL32"
                };

//
// Dump the base relocation table of a PE file
//
template <class T>
static void DumpBaseRelocationsSection (const BYTE *imageBase, const T *pNTHeader)
{
    const IMAGE_BASE_RELOCATION *baseReloc;
    DWORD dwBaseRelocRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_BASERELOC);

    if (!dwBaseRelocRVA)
       return;

    baseReloc = (const IMAGE_BASE_RELOCATION*)
                GetPtrFromRVA (dwBaseRelocRVA, pNTHeader, imageBase);
    if (!baseReloc)
       return;

    printf ("base relocations:\n\n");

    while (baseReloc->SizeOfBlock)
    {
        const WORD *pEntry;
        const char *szRelocType;
        unsigned    i, cEntries;
        WORD        relocType;

        // Sanity check to make sure the data looks OK.
        if (baseReloc->VirtualAddress == 0)
           break;
        if (baseReloc->SizeOfBlock < sizeof(*baseReloc))
           break;

        cEntries = (baseReloc->SizeOfBlock - sizeof(*baseReloc)) / sizeof(WORD);
        pEntry = MakePtr (const WORD*, baseReloc, sizeof(*baseReloc));

        printf ("Virtual Address: %08lX  size: %08lX\n",
                baseReloc->VirtualAddress, baseReloc->SizeOfBlock);

        for (i = 0; i < cEntries; i++)
        {
            // Extract the top 4 bits of the relocation entry.  Turn those 4
            // bits into an appropriate descriptive string (szRelocType)
            relocType = (*pEntry & 0xF000) >> 12;
            szRelocType = relocType <= IMAGE_REL_BASED_DIR64 ? SzRelocTypes[relocType] : "unknown";

            printf ("  %08lX %s",
                    (*pEntry & 0x0FFF) + baseReloc->VirtualAddress, szRelocType);

            if (IMAGE_REL_BASED_HIGHADJ == relocType)
            {
              pEntry++;
              cEntries--;
              printf (" (%X)", *pEntry);
            }
            printf ("\n");
            pEntry++;   // Advance to next relocation entry
        }
        baseReloc = MakePtr (const IMAGE_BASE_RELOCATION *, baseReloc,
                             baseReloc->SizeOfBlock);
    }
}

//
// Dump out the new IMAGE_BOUND_IMPORT_DESCRIPTOR that NT 3.51 added
//
template <class T>
static void DumpBoundImportDescriptors (const BYTE *imageBase,
                                        const T    *pNTHeader)
{
    const IMAGE_BOUND_IMPORT_DESCRIPTOR *pibid;
    DWORD bidRVA = GetImgDirEntryRVA(pNTHeader, IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT);

    if (!bidRVA) // Bound import descriptors RVA
        return;

    pibid = MakePtr (PIMAGE_BOUND_IMPORT_DESCRIPTOR, imageBase, bidRVA);

    printf ("Bound import descriptors:\n\n");
    printf ("  Module        TimeDate\n");
    printf ("  ------------  --------\n");

    while (pibid->TimeDateStamp)
    {
        const IMAGE_BOUND_FORWARDER_REF *pibfr;
        DWORD i;

        time_t timeStamp = pibid->TimeDateStamp;

        printf ("  %-12s  %08lX -> %s",
                imageBase + bidRVA + pibid->OffsetModuleName,
                pibid->TimeDateStamp,
                ctime(&timeStamp));

        pibfr = MakePtr (const IMAGE_BOUND_FORWARDER_REF*, pibid,
                         sizeof(IMAGE_BOUND_IMPORT_DESCRIPTOR));

        for (i = 0; i < pibid->NumberOfModuleForwarderRefs; i++, pibfr++)
        {
            timeStamp = pibfr->TimeDateStamp;

            printf ("    forwarder:  %-12s  %08lX -> %s",
                    imageBase + bidRVA + pibfr->OffsetModuleName,
                    pibfr->TimeDateStamp, ctime(&timeStamp));

            // Keep the outer loop pointer up to date too!
            pibid = MakePtr (const IMAGE_BOUND_IMPORT_DESCRIPTOR *, pibid,
                             sizeof(IMAGE_BOUND_FORWARDER_REF));
        }
        pibid++;    // Advance to next pibid;
    }
    printf ("\n");
}


template <class T>
static void DumpCOR20Header (const BYTE *imageBase, const T *pNTHeader)
{
    const IMAGE_COR20_HEADER *pCor20Hdr;
    DWORD cor20HdrRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR);

    if (!cor20HdrRVA)
        return;

    pCor20Hdr = (const IMAGE_COR20_HEADER*) GetPtrFromRVA (cor20HdrRVA, pNTHeader, imageBase);

    printf (".NET Runtime Header:\n");

    printf ("  Size:       %lu\n", pCor20Hdr->cb);
    printf ("  Version:    %u.%u\n", pCor20Hdr->MajorRuntimeVersion, pCor20Hdr->MinorRuntimeVersion);
    printf ("  Flags:      %lX\n", pCor20Hdr->Flags);

    if (pCor20Hdr->Flags & COMIMAGE_FLAGS_ILONLY)
       printf ("    ILONLY\n");

    if (pCor20Hdr->Flags & COMIMAGE_FLAGS_32BITREQUIRED)
       printf ("    32BITREQUIRED\n");

    if (pCor20Hdr->Flags & COMIMAGE_FLAGS_IL_LIBRARY)
       printf ("    IL_LIBRARY\n");

    if (pCor20Hdr->Flags & 8)
       printf ("    STRONGNAMESIGNED\n");     // At this moment, WINNT.H and CorHdr.H are out of sync...

    if (pCor20Hdr->Flags & COMIMAGE_FLAGS_TRACKDEBUGDATA)
       printf ("    TRACKDEBUGDATA\n");

    DisplayDataDirectoryEntry ("MetaData", &pCor20Hdr->MetaData);
    DisplayDataDirectoryEntry ("Resources", &pCor20Hdr->Resources);
    DisplayDataDirectoryEntry ("StrongNameSig", &pCor20Hdr->StrongNameSignature);
    DisplayDataDirectoryEntry ("CodeManagerTable", &pCor20Hdr->CodeManagerTable);
    DisplayDataDirectoryEntry ("VTableFixups", &pCor20Hdr->VTableFixups);
    DisplayDataDirectoryEntry ("ExprtAddrTblJmps", &pCor20Hdr->ExportAddressTableJumps);
    DisplayDataDirectoryEntry ("ManagedNativeHdr", &pCor20Hdr->ManagedNativeHeader);

    printf ("\n");
}

static void DumpLoadConfigDirectory32 (const BYTE               *imageBase,
                                       const IMAGE_NT_HEADERS32 *pNTHeader)
{
    const IMAGE_LOAD_CONFIG_DIRECTORY32* pLCD;
    DWORD loadConfigDirRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);

    if (!loadConfigDirRVA)
        return;

    pLCD = (const IMAGE_LOAD_CONFIG_DIRECTORY32*) GetPtrFromRVA (loadConfigDirRVA, pNTHeader, imageBase);

    printf ("Image Load Configuration Directory:\n");

//  printf ("  Characteristics:               %X\n", pLCD->Characteristics);
    printf ("  TimeDateStamp:                 %lX\n", pLCD->TimeDateStamp);
    printf ("  Version:                       %u.%u\n", pLCD->MajorVersion, pLCD->MinorVersion);
    printf ("  GlobalFlagsClear:              %lX\n", pLCD->GlobalFlagsClear);
    printf ("  GlobalFlagsSet:                %lX\n", pLCD->GlobalFlagsSet);
    printf ("  CriticalSectionDefaultTimeout: %lX\n", pLCD->CriticalSectionDefaultTimeout);
    printf ("  DeCommitFreeBlockThreshold:    %lX\n", pLCD->DeCommitFreeBlockThreshold);
    printf ("  DeCommitTotalFreeThreshold:    %lX\n", pLCD->DeCommitTotalFreeThreshold);
    printf ("  LockPrefixTable:               %lX\n", (DWORD)pLCD->LockPrefixTable);
    printf ("  MaximumAllocationSize:         %lX\n", pLCD->MaximumAllocationSize);
    printf ("  VirtualMemoryThreshold:        %lX\n", pLCD->VirtualMemoryThreshold);
    printf ("  ProcessHeapFlags:              %lX\n", pLCD->ProcessHeapFlags);
#if !defined(__MINGW32__)
    printf ("  ProcessAffinityMask:           %lX\n", pLCD->ProcessAffinityMask);
    printf ("  CSDVersion:                    %u\n",  pLCD->CSDVersion);
//  printf ("  Reserved1:                     %lX\n", pLCD->Reserved1);
    printf ("  EditList:                      %lX\n", pLCD->EditList);
#endif

   printf ("\n");
}

static void DumpLoadConfigDirectory64 (const BYTE               *imageBase,
                                       const IMAGE_NT_HEADERS64 *pNTHeader)
{
    const IMAGE_LOAD_CONFIG_DIRECTORY64* pLCD;
    DWORD loadConfigDirRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);

    if (!loadConfigDirRVA)
        return;

    pLCD = (const IMAGE_LOAD_CONFIG_DIRECTORY64*) GetPtrFromRVA (loadConfigDirRVA, pNTHeader, imageBase);

    printf ("Image Load Configuration Directory:\n");

//  printf ("  Characteristics:               %X\n", pLCD->Characteristics);
    printf ("  TimeDateStamp:                 %lX\n", pLCD->TimeDateStamp);
    printf ("  Version:                       %u.%u\n", pLCD->MajorVersion, pLCD->MinorVersion);
    printf ("  GlobalFlagsClear:              %lX\n", pLCD->GlobalFlagsClear);
    printf ("  GlobalFlagsSet:                %lX\n", pLCD->GlobalFlagsSet);
    printf ("  CriticalSectionDefaultTimeout: %lX\n", pLCD->CriticalSectionDefaultTimeout);
    printf ("  DeCommitFreeBlockThreshold:    %I64X\n", pLCD->DeCommitFreeBlockThreshold);
    printf ("  DeCommitTotalFreeThreshold:    %I64X\n", pLCD->DeCommitTotalFreeThreshold);
    printf ("  LockPrefixTable:               %I64X\n", pLCD->LockPrefixTable);
    printf ("  MaximumAllocationSize:         %I64X\n", pLCD->MaximumAllocationSize);
    printf ("  VirtualMemoryThreshold:        %I64X\n", pLCD->VirtualMemoryThreshold);
    printf ("  ProcessHeapFlags:              %lX\n", pLCD->ProcessHeapFlags);
#if !defined(__MINGW32__)
    printf ("  ProcessAffinityMask:           %llX\n", pLCD->ProcessAffinityMask);
    printf ("  CSDVersion:                    %u\n", pLCD->CSDVersion);
//  printf ("  Reserved1:                     %X\n", pLCD->Reserved1);
    printf ("  EditList:                      %I64X\n", pLCD->EditList);
#endif

   printf ("\n");
}

template <class T>
static void DumpCertificates (const BYTE *imageBase, const T *pNTHeader)
{
    // Note that the this DataDirectory entry gives a >>> FILE OFFSET <<< rather than
    // an RVA.
    DWORD dwTotalSize, certOffset = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_SECURITY);

    if (!certOffset)
       return;

    dwTotalSize = GetImgDirEntrySize (pNTHeader, IMAGE_DIRECTORY_ENTRY_SECURITY);

    printf ("Certificates:\n");

    while (dwTotalSize)   // As long as there is unprocessed certificate data...
    {
        const WIN_CERTIFICATE *pCert = MakePtr (const WIN_CERTIFICATE *, imageBase, certOffset);

        printf ("  Length:   %lX\n", pCert->dwLength);
        printf ("  Revision: %X\n", pCert->wRevision);
        printf ("  Type:     %X", pCert->wCertificateType);

        switch (pCert->wCertificateType)
        {
          case WIN_CERT_TYPE_X509:
               printf (" (X509)");
               break;
          case WIN_CERT_TYPE_PKCS_SIGNED_DATA:
               printf (" (PKCS_SIGNED_DATA)");
               break;
          case WIN_CERT_TYPE_TS_STACK_SIGNED:
               printf (" (TS_STACK_SIGNED)");
               break;
        }
#if 0
        hex_dump (&pCert->bCertificate, pCert->dwLength);
#endif
        printf ("\n");

        dwTotalSize -= pCert->dwLength;
        certOffset  += pCert->dwLength;      // Get offset to next certificate
    }
    printf ("\n");
}

/*
 * Verify the checksum of last opened file above.
 */
template <class T>
static BOOL verify_checksum (const T *optionalHeader)
{
  DWORD header_sum, calc_chk_sum;

  if (!p_CheckSumMappedFile)
     return (FALSE);

  (*p_CheckSumMappedFile) ((void*)g_mmap_start, g_file_size.LowPart,
                           &header_sum, &calc_chk_sum);
  return (calc_chk_sum == optionalHeader->CheckSum);
}


//
// Dump the IMAGE_OPTIONAL_HEADER from a PE file
//
template <class T>
static void DumpOptionalHeader (const T *optionalHeader)
{
    const char *s;
    UINT  i, width = 30;

    printf ("Optional Header\n");

    printf ("  %-*s%04X\n", width, "Magic", optionalHeader->Magic);
    printf ("  %-*s%u.%02u\n", width, "linker version",
            optionalHeader->MajorLinkerVersion, optionalHeader->MinorLinkerVersion);

    printf ("  %-*s%lX\n", width, "size of code",
            optionalHeader->SizeOfCode);

    printf ("  %-*s%lX\n", width, "size of initialized data",
            optionalHeader->SizeOfInitializedData);

    printf ("  %-*s%lX\n", width, "size of uninitialized data",
            optionalHeader->SizeOfUninitializedData);

    printf ("  %-*s%lX\n", width, "entrypoint RVA",
            optionalHeader->AddressOfEntryPoint);

    printf ("  %-*s%lX\n", width, "base of code", optionalHeader->BaseOfCode);

#if 0
    BOOL b64BitHeader = (IMAGE_NT_OPTIONAL_HDR64_MAGIC == optionalHeader->Magic);
    size_t s_T = sizeof(T);

    if (!b64BitHeader)
       printf ("  %-*s%X\n", width, "base of data", optionalHeader->BaseOfData);
#endif

    printf ("  %-*s%I64X\n", width, "image base", (ULONGLONG)optionalHeader->ImageBase);

    printf ("  %-*s%lX\n", width, "section align",
            optionalHeader->SectionAlignment);

    printf ("  %-*s%lX\n", width, "file align",
            optionalHeader->FileAlignment);

    printf ("  %-*s%d.%02d\n", width, "required OS version",
            optionalHeader->MajorOperatingSystemVersion,
            optionalHeader->MinorOperatingSystemVersion);

    printf ("  %-*s%u.%02u\n", width, "image version",
            optionalHeader->MajorImageVersion,
            optionalHeader->MinorImageVersion);

    printf ("  %-*s%u.%02u\n", width, "subsystem version",
            optionalHeader->MajorSubsystemVersion,
            optionalHeader->MinorSubsystemVersion);

    printf ("  %-*s%lX\n", width, "Win32 Version",
            optionalHeader->Win32VersionValue);

    printf ("  %-*s%lX\n", width, "size of image",
            optionalHeader->SizeOfImage);

    printf ("  %-*s%lX\n", width, "size of headers",
            optionalHeader->SizeOfHeaders);

    printf ("  %-*s%lX  (%s)\n", width, "checksum",
            optionalHeader->CheckSum,
            verify_checksum(optionalHeader) ? "OK" : "Failed");

    switch (optionalHeader->Subsystem)
    {
        case IMAGE_SUBSYSTEM_NATIVE:
             s = "Native";
             break;
        case IMAGE_SUBSYSTEM_WINDOWS_GUI:
             s = "Windows GUI";
             break;
        case IMAGE_SUBSYSTEM_WINDOWS_CUI:
             s = "Windows character";
             break;
        case IMAGE_SUBSYSTEM_OS2_CUI:
             s = "OS/2 character";
             break;
        case IMAGE_SUBSYSTEM_POSIX_CUI:
             s = "Posix character";
             break;
        default:
             s = "unknown";
             break;
    }
    printf ("  %-*s%04X (%s)\n", width, "Subsystem",
            optionalHeader->Subsystem, s);

// Marked as obsolete in MSDN CD 9
    printf ("  %-*s%04X\n", width, "DLL flags",
            optionalHeader->DllCharacteristics);

    for (i = 0; i < NUMBER_DLL_CHARACTERISTICS; i++)
    {
      if (optionalHeader->DllCharacteristics & DllCharacteristics[i].flag)
         printf ("  %-*s%s", width, " ", DllCharacteristics[i].name );
    }

    if (optionalHeader->DllCharacteristics)
       printf ("\n");

    printf ("  %-*s%lX\n", width, "stack reserve size",
            (DWORD)optionalHeader->SizeOfStackReserve);

    printf ("  %-*s%lX\n", width, "stack commit size",
            (DWORD)optionalHeader->SizeOfStackCommit);

    printf ("  %-*s%lX\n", width, "heap reserve size",
            (DWORD)optionalHeader->SizeOfHeapReserve);

    printf ("  %-*s%lX\n", width, "heap commit size",
            (DWORD)optionalHeader->SizeOfHeapCommit);

#if 0   // Marked as obsolete in MSDN CD 9
    printf ("  %-*s%08X\n", width, "loader flags",
            optionalHeader->LoaderFlags);

    for ( i = 0; i < NUMBER_LOADER_FLAGS; i++ )
    {
        if (optionalHeader->LoaderFlags & LoaderFlags[i].flag)
            printf ("  %s", LoaderFlags[i].name);
    }
    if (optionalHeader->LoaderFlags)
       printf ("\n");
#endif

    printf ("  %-*s%lX\n", width, "RVAs & sizes",
            optionalHeader->NumberOfRvaAndSizes);

    printf ("\nData Directory\n");
    for (i = 0; i < optionalHeader->NumberOfRvaAndSizes; i++)
    {
        printf ("  %-12s rva: %08lX  size: %08lX\n",
                (i >= NUMBER_IMAGE_DIRECTORY_ENTRYS) ? "unused" : ImageDirectoryNames[i],
                optionalHeader->DataDirectory[i].VirtualAddress,
                optionalHeader->DataDirectory[i].Size);
    }
}

//
// top level routine called from PEDUMP.C to dump the components of a PE file
//
void DumpExeFile (const IMAGE_DOS_HEADER *dosHeader)
{
    const IMAGE_NT_HEADERS32 *pNTHeader32;
    const IMAGE_NT_HEADERS64 *pNTHeader64;

    const BYTE *base = (const BYTE*) dosHeader;

    pNTHeader32 = MakePtr (const IMAGE_NT_HEADERS *, dosHeader, dosHeader->e_lfanew);
    pNTHeader64 = (const IMAGE_NT_HEADERS64 *) pNTHeader32;

    // First, verify that the e_lfanew field gave us a reasonable
    // pointer, then verify the PE signature.

    if (!PTR_IN_RANGE (pNTHeader32, sizeof(*pNTHeader32)) ||
        pNTHeader32->Signature != IMAGE_NT_SIGNATURE)
    {
      printf ("Not a Portable Executable (PE) EXE\n");
      return;
    }

    DumpHeader (&pNTHeader32->FileHeader);
    printf ("\n");

    bool bIs64Bit = (pNTHeader32->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC);

    if (bIs64Bit)
         DumpOptionalHeader (&pNTHeader64->OptionalHeader);
    else DumpOptionalHeader (&pNTHeader32->OptionalHeader);

    printf ("\n");

    DumpSectionTable (IMAGE_FIRST_SECTION(pNTHeader32),
                      pNTHeader32->FileHeader.NumberOfSections, TRUE);
    printf ("\n");

    if (bIs64Bit)
         DumpExeDebugDirectory (base, pNTHeader64);
    else DumpExeDebugDirectory (base, pNTHeader32);

    if (pNTHeader32->FileHeader.PointerToSymbolTable == 0)
        g_pCOFFHeader = 0; // Doesn't really exist!
    printf ("\n");

    DumpResources (base, pNTHeader32);
    printf ("\n");

    bIs64Bit
        ? DumpTLSDirectory64 (base, pNTHeader64)
        : DumpTLSDirectory32 (base, pNTHeader32);

    bIs64Bit
        ? DumpImportsSection(base, pNTHeader64)
        : DumpImportsSection(base, pNTHeader32);

    bIs64Bit
        ? DumpDelayedImportsSection(base, pNTHeader64, bIs64Bit)
        : DumpDelayedImportsSection(base, pNTHeader32, bIs64Bit);


    bIs64Bit
        ? DumpBoundImportDescriptors (base, pNTHeader64)
        : DumpBoundImportDescriptors (base, pNTHeader32);

    bIs64Bit
        ? DumpExportsSection (base, pNTHeader64)
        : DumpExportsSection (base, pNTHeader32);

    bIs64Bit
        ? DumpCOR20Header (base, pNTHeader64)
        : DumpCOR20Header (base, pNTHeader32);

    bIs64Bit
        ? DumpLoadConfigDirectory64 (base, pNTHeader64)
        : DumpLoadConfigDirectory32 (base, pNTHeader32);

    bIs64Bit
        ? DumpCertificates (base, pNTHeader64)
        : DumpCertificates (base, pNTHeader32);

    //=========================================================================
    //
    // If we have COFF symbols, create a symbol table now
    //
    //=========================================================================

    if (g_pCOFFHeader)    // Did we see a COFF symbols header while looking
    {                       // through the debug directory?
        g_pCOFFSymbolTable = new COFFSymbolTable(
                (PVOID)(base + pNTHeader32->FileHeader.PointerToSymbolTable),
                pNTHeader32->FileHeader.NumberOfSymbols);
    }

    if (fShowPDATA)
    {
        bIs64Bit
            ? DumpRuntimeFunctions (base, pNTHeader64)
            : DumpRuntimeFunctions (base, pNTHeader32);
        printf ("\n");
    }

    if (fShowRelocations)
    {
        bIs64Bit
            ? DumpBaseRelocationsSection (base, pNTHeader64)
            : DumpBaseRelocationsSection (base, pNTHeader32);
        printf ("\n");
    }

    if (fShowSymbolTable && g_pMiscDebugInfo)
    {
        DumpMiscDebugInfo (g_pMiscDebugInfo);
        printf ("\n");
    }

    if (fShowSymbolTable && g_pCVHeader)
    {
        DumpCVDebugInfoRecord (g_pCVHeader);
        printf ("\n");
    }

    if (fShowSymbolTable && g_pCOFFHeader)
    {
        DumpCOFFHeader (g_pCOFFHeader);
        printf ("\n");
    }

    if (fShowLineNumbers && g_pCOFFHeader)
    {
        DumpLineNumbers (MakePtr(const IMAGE_LINENUMBER*, g_pCOFFHeader,
                            g_pCOFFHeader->LvaToFirstLinenumber),
                            g_pCOFFHeader->NumberOfLinenumbers);
        printf ("\n");
    }

    if (fShowSymbolTable)
    {
        if (pNTHeader32->FileHeader.NumberOfSymbols &&
            pNTHeader32->FileHeader.PointerToSymbolTable &&
            g_pCOFFSymbolTable)
        {
            DumpCOFFSymbolTable (g_pCOFFSymbolTable);
            printf ("\n");
        }
    }

    if (fShowRawSectionData)
    {
        const IMAGE_SECTION_HEADER *pSectionHdr;

        pSectionHdr = bIs64Bit ? (const IMAGE_SECTION_HEADER*)(pNTHeader64+1) :
                                 (const IMAGE_SECTION_HEADER*)(pNTHeader32+1);

        DumpRawSectionData (pSectionHdr, dosHeader, pNTHeader32->FileHeader.NumberOfSections);
    }

    if (g_pCOFFSymbolTable)
        delete g_pCOFFSymbolTable;
}
