//==================================
// PEDUMP - Matt Pietrek 1997-2001
// FILE: RESDUMP.CPP
//==================================

#include "Common.h"
#include "ExtrnVar.h"

#ifndef min
#define min(a,b)       (((a) < (b)) ? (a) : (b))
#endif

#define RES2WORD(res)  (WORD) ( (DWORD_PTR)(res) & 0xFFFF )

static const IMAGE_RESOURCE_DIRECTORY_ENTRY *g_pStrResEntries = NULL;
static const IMAGE_RESOURCE_DIRECTORY_ENTRY *g_pDlgResEntries = NULL;
static DWORD                                 g_cStrResEntries = 0;
static DWORD                                 g_cDlgResEntries = 0;

static void DumpResourceDirectory (const IMAGE_RESOURCE_DIRECTORY *pResDir,
                                   const BYTE *pResourceBase,
                                   DWORD level, DWORD resourceType);


static DWORD GetOffsetToDataFromResEntry (const BYTE *pResourceBase,
                                          const IMAGE_RESOURCE_DIRECTORY_ENTRY *pResEntry)
{
  // The IMAGE_RESOURCE_DIRECTORY_ENTRY is gonna point to a single
  // IMAGE_RESOURCE_DIRECTORY, which in turn will point to the
  // IMAGE_RESOURCE_DIRECTORY_ENTRY, which in turn will point
  // to the IMAGE_RESOURCE_DATA_ENTRY that we're really after.  In
  // other words, traverse down a level.

  const IMAGE_RESOURCE_DIRECTORY       *pStupidResDir = (const IMAGE_RESOURCE_DIRECTORY*) (pResourceBase + pResEntry->OffsetToDirectory);
  const IMAGE_RESOURCE_DIRECTORY_ENTRY *pResDirEntry  = (const IMAGE_RESOURCE_DIRECTORY_ENTRY*) (pStupidResDir + 1);
  const IMAGE_RESOURCE_DATA_ENTRY      *pResDataEntry = (const IMAGE_RESOURCE_DATA_ENTRY*) (pResourceBase + pResDirEntry->OffsetToData);

  return (pResDataEntry->OffsetToData);
}

// Get an ASCII string representing a resource type
static void GetResourceTypeName (DWORD type, char *buffer, UINT cBytes)
{
  // The predefined resource types
  static const char *resourceTypes[] = {
    "???_0",
    "CURSOR",
    "BITMAP",
    "ICON",
    "MENU",
    "DIALOG",
    "STRING",
    "FONTDIR",
    "FONT",
    "ACCELERATORS",
    "RCDATA",
    "MESSAGETABLE",
    "GROUP_CURSOR",
    "???_13",
    "GROUP_ICON",
    "???_15",
    "VERSION",
    "DLGINCLUDE",
    "???_18",
    "PLUGPLAY",
    "VXD",
    "ANICURSOR",
    "ANIICON",
    "???_23",
    "MANIFEST"
  };

  if (type <= DIM(resourceTypes))
       strncpy (buffer, resourceTypes[type], cBytes);
  else sprintf (buffer, "%lX", type);
}

//
// If a resource entry has a string name (rather than an ID), go find
// the string and convert it from unicode to ascii.
//
static void GetResourceNameFromId (DWORD id, const BYTE *pResourceBase, char *buffer, UINT cBytes)
{
  const IMAGE_RESOURCE_DIR_STRING_U *prdsu;

  // If it's a regular ID, just format it.
  if (!(id & IMAGE_RESOURCE_NAME_IS_STRING))
  {
    sprintf (buffer, "%lX", id);
    return;
  }

  id &= 0x7FFFFFFF;
  prdsu = MakePtr (const IMAGE_RESOURCE_DIR_STRING_U *, pResourceBase, id);

  // prdsu->Length is the number of unicode characters
  WideCharToMultiByte (CP_ACP, 0, prdsu->NameString, prdsu->Length, buffer, cBytes, 0, 0);
  buffer [min(cBytes - 1, prdsu->Length)] = 0; // Null terminate it!!!
}

//
// Dump the information about one resource directory entry.  If the
// entry is for a subdirectory, call the directory dumping routine
// instead of printing information in this routine.
//
static void DumpResourceEntry (const IMAGE_RESOURCE_DIRECTORY_ENTRY *pResDirEntry,
                               const BYTE *pResourceBase, DWORD level)
{
  const IMAGE_RESOURCE_DATA_ENTRY *pResDataEntry;
  char nameBuffer[128];
  UINT i;

  if (pResDirEntry->OffsetToData & IMAGE_RESOURCE_DATA_IS_DIRECTORY)
  {
    DumpResourceDirectory ((const IMAGE_RESOURCE_DIRECTORY*)
                           ((pResDirEntry->OffsetToData & 0x7FFFFFFF) + pResourceBase),
                           pResourceBase, level, pResDirEntry->Name);
    return;
  }

  // Spit out the spacing for the level indentation
  for (i = 0; i < level; i++)
    printf ("    ");

  if (pResDirEntry->Name & IMAGE_RESOURCE_NAME_IS_STRING)
  {
    GetResourceNameFromId (pResDirEntry->Name, pResourceBase, nameBuffer, sizeof (nameBuffer));
    printf ("Name: %s  DataEntryOffs: %08lX\n", nameBuffer, pResDirEntry->OffsetToData);
  }
  else
  {
    printf ("ID: %08lX  DataEntryOffs: %08lX\n", pResDirEntry->Name, pResDirEntry->OffsetToData);
  }

  // the resDirEntry->OffsetToData is a pointer to an
  // IMAGE_RESOURCE_DATA_ENTRY.  Go dump out that information.  First,
  // spit out the proper indentation
  for (i = 0; i < level; i++)
    printf ("    ");

  pResDataEntry = MakePtr (const IMAGE_RESOURCE_DATA_ENTRY *, pResourceBase, pResDirEntry->OffsetToData);
  printf ("DataRVA: %05lX  DataSize: %05lX  CodePage: %lX\n",
          pResDataEntry->OffsetToData, pResDataEntry->Size, pResDataEntry->CodePage);
}

//
// Dump the information about one resource directory.
//
static void DumpResourceDirectory (const IMAGE_RESOURCE_DIRECTORY *pResDir,
                                   const BYTE *pResourceBase,
                                   DWORD level, DWORD resourceType)
{
  const IMAGE_RESOURCE_DIRECTORY_ENTRY *resDirEntry;
  char szType[64];
  UINT i;

  // Level 1 resources are the resource types
  if (level == 1)
  {
    printf ("    ---------------------------------------------------" "-----------\n");

    if (resourceType & IMAGE_RESOURCE_NAME_IS_STRING)
    {
      GetResourceNameFromId (resourceType, pResourceBase, szType, sizeof (szType));
    }
    else
    {
      GetResourceTypeName (resourceType, szType, sizeof (szType));
    }
  }
  else                          // All other levels, just print out the regular id or name
  {
    GetResourceNameFromId (resourceType, pResourceBase, szType, sizeof (szType));
  }

  // Spit out the spacing for the level indentation
  for (i = 0; i < level; i++)
    printf ("    ");

  printf ("ResDir (%s) Entries:%02X (Named:%02X, ID:%02X) \0 TimeDate:%08lX",
          szType, pResDir->NumberOfNamedEntries + pResDir->NumberOfIdEntries,
          pResDir->NumberOfNamedEntries, pResDir->NumberOfIdEntries /* , pResDir->TimeDateStamp */);

  if (pResDir->MajorVersion || pResDir->MinorVersion)
     printf (" Vers:%u.%02u", pResDir->MajorVersion, pResDir->MinorVersion);

  if (pResDir->Characteristics)
     printf (" Char:%08lX", pResDir->Characteristics);

  printf ("\n");

  //
  // The "directory entries" immediately follow the directory in memory
  //
  resDirEntry = (const IMAGE_RESOURCE_DIRECTORY_ENTRY*) (pResDir + 1);

  // If it's a stringtable, save off info for future use
  if (level == 1 && resourceType == RES2WORD(RT_STRING))
  {
    g_pStrResEntries = (const IMAGE_RESOURCE_DIRECTORY_ENTRY*) resDirEntry;
    g_cStrResEntries = pResDir->NumberOfIdEntries;
  }

  // If it's a stringtable, save off info for future use
  if (level == 1 && resourceType == RES2WORD(RT_DIALOG))
  {
    g_pDlgResEntries = (const IMAGE_RESOURCE_DIRECTORY_ENTRY*) resDirEntry;
    g_cDlgResEntries = pResDir->NumberOfIdEntries;
  }

  for (i = 0; i < pResDir->NumberOfNamedEntries; i++, resDirEntry++)
    DumpResourceEntry (resDirEntry, pResourceBase, level + 1);

  for (i = 0; i < pResDir->NumberOfIdEntries; i++, resDirEntry++)
    DumpResourceEntry (resDirEntry, pResourceBase, level + 1);
}

#if 0
  #define PRINTABLE(c) isprint(c)
#else
  #define PRINTABLE(c) ((c) >= ' ')
#endif

#define MAX_RES_STRING  80

template <class T>
static void DumpStringTable (const BYTE                           *pImageBase,
                             const T                              *pNTHeader,
                             const BYTE                           *pResourceBase,
                             const IMAGE_RESOURCE_DIRECTORY_ENTRY *pStrResEntry,
                             DWORD                                 cStrResEntries)
{
  for (unsigned i = 0; i < cStrResEntries; i++, pStrResEntry++)
  {
    DWORD offsetToData = GetOffsetToDataFromResEntry (pResourceBase, pStrResEntry);

    const WORD *pStrEntry = (const WORD*) GetPtrFromRVA (offsetToData,
                                                         pNTHeader, pImageBase);
    if (!pStrEntry)
       break;

    unsigned id = (pStrResEntry->Name - 1) << 4;

    for (unsigned j = 0; j < 16; j++)
    {
      WORD k, len = *pStrEntry++;

      if (len)
      {
#if 0
        printf ("%-5u: ", id + j);
#else
        printf ("  %04X: ", id + j);
#endif
        for (k = 0; k < min(len,MAX_RES_STRING); k++)
        {
          const char *s;
          char  szBuff[20];
          BYTE  c = (BYTE) pStrEntry[k];

          switch (c)
          {
            case '\t':
                 s = "\\t";
                 break;
            case '\r':
                 s = "\\r";
                 break;
            case '\n':
                 s = "\\n";
                 break;
            default:
                 sprintf (szBuff, "%c", PRINTABLE(c) ? c : '.');
                 s = szBuff;
                 break;
          }
          printf (s);
        }
        if (k && k < len)
           printf ("...");
        printf ("\n");
      }
      pStrEntry += len;
    }
  }
}

//============================================================================

template <class T>
static void DumpDialogs (const BYTE                           *pImageBase,
                         const T                              *pNTHeader,
                         const BYTE                           *pResourceBase,
                         const IMAGE_RESOURCE_DIRECTORY_ENTRY *pDlgResEntry,
                         DWORD                                 cDlgResEntries)
{
  for (unsigned i = 0; i < cDlgResEntries; i++, pDlgResEntry++)
  {
    DWORD offsetToData = GetOffsetToDataFromResEntry (pResourceBase, pDlgResEntry);

    const DWORD *pDlgStyle = (const DWORD*) GetPtrFromRVA (offsetToData,
                                                           pNTHeader, pImageBase);

    if (!pDlgStyle)
      break;

    printf ("  ====================\n");

    if (HIWORD(*pDlgStyle) != 0xFFFF)
    {
      // A regular DLGTEMPLATE
      const DLGTEMPLATE *pDlgTemplate = (const DLGTEMPLATE *) pDlgStyle;

      printf ("  style: %08lX\n", pDlgTemplate->style);
      printf ("  extended style: %08lX\n", pDlgTemplate->dwExtendedStyle);

      printf ("  controls: %u\n", pDlgTemplate->cdit);
      printf ("  (%u,%u) - (%u,%u)\n",
              pDlgTemplate->x, pDlgTemplate->y,
              pDlgTemplate->x + pDlgTemplate->cx,
              pDlgTemplate->y + pDlgTemplate->cy);

      const WORD *pMenu = (const WORD*) (pDlgTemplate + 1);

      //
      // First comes the menu
      //
      if (*pMenu)
      {
        if (*pMenu == 0xFFFF)
        {
          pMenu++;
          printf ("  ordinal menu: %u\n", *pMenu);
        }
        else
        {
          printf ("  menu: ");
          while (*pMenu)
            printf ("%c", LOBYTE(*pMenu++));
          pMenu++;
          printf ("\n");
        }
      }
      else
        pMenu++;                // Advance past the menu name

      //
      // Next comes the class
      //
      const WORD *pClass = pMenu;

      if (*pClass)
      {
        if (*pClass == 0xFFFF)
        {
          pClass++;
          printf ("  ordinal class: %u\n", *pClass);
        }
        else
        {
          printf ("  class: ");
          while (*pClass)
            printf ("%c", LOBYTE(*pClass++));
          pClass++;
          printf ("\n");
        }
      }
      else
        pClass++;               // Advance past the class name

      //
      // Finally comes the title
      //
      const WORD *pTitle = pClass;

      if (*pTitle)
      {
        printf ("  title: ");
        while (*pTitle)
          printf ("%c", LOBYTE(*pTitle++));
        pTitle++;
      }
      else
        pTitle++;               // Advance past the Title name

      printf ("\n");

      const WORD *pFont = pTitle;

      if (pDlgTemplate->style & DS_SETFONT)
      {
        printf ("  Font: %u point ", *pFont++);
        while (*pFont)
          printf ("%c", LOBYTE(*pFont++));
        pFont++;
        printf ("\n");
      }
      else
        pFont = pTitle;


      // DLGITEMPLATE starts on a 4 byte boundary
      const DLGITEMTEMPLATE *pDlgItemTemplate = (const DLGITEMTEMPLATE*) pFont;

      for (unsigned i = 0; i < pDlgTemplate->cdit; i++)
      {
        // Control item header....
        pDlgItemTemplate = (const DLGITEMTEMPLATE *) (((DWORD_PTR) pDlgItemTemplate + 3) & ~3);

        printf ("    style:          %08lX\n", pDlgItemTemplate->style);
        printf ("    extended style: %08lX\n", pDlgItemTemplate->dwExtendedStyle);

        printf ("    (%u,%u) - (%u,%u)\n",
                pDlgItemTemplate->x, pDlgItemTemplate->y,
                pDlgItemTemplate->x + pDlgItemTemplate->cx, pDlgItemTemplate->y + pDlgItemTemplate->cy);
        printf ("    id: %u\n", pDlgItemTemplate->id);

        //
        // Next comes the control's class name or ID
        //
        const WORD *pClass = (const WORD*) (pDlgItemTemplate + 1);

        if (*pClass)
        {
          if (*pClass == 0xFFFF)
          {
            pClass++;
            printf ("    ordinal class: %u", *pClass++);
          }
          else
          {
            printf ("    class: ");
            while (*pClass)
              printf ("%c", LOBYTE(*pClass++));

            pClass++;
            printf ("\n");
          }
        }
        else
          pClass++;

        printf ("\n");

        //
        // next comes the title
        //
        const WORD *pTitle = pClass;

        if (*pTitle)
        {
          printf ("    title: ");
          if (*pTitle == 0xFFFF)
          {
            pTitle++;
            printf ("%u\n", *pTitle++);
          }
          else
          {
            while (*pTitle)
              printf ("%c", LOBYTE(*pTitle++));
            pTitle++;
            printf ("\n");
          }
        }
        else
          pTitle++;             // Advance past the Title name

        printf ("\n");

        // Round up or down to 2 byte boundary (?)
        const BYTE *pCreationData = (const BYTE*) (((DWORD_PTR) pTitle + 1) & (0 - 2));

        if (*pCreationData)
             pCreationData += *pCreationData;
        else pCreationData++;

        pDlgItemTemplate = (const DLGITEMTEMPLATE *) pCreationData;
        printf ("\n");
      }
      printf ("\n");
    }
    else
    {
      // A DLGTEMPLATEEX
    }

    printf ("\n");
  }
}


//
// Top level routine called to dump out the entire resource hierarchy
//
template <class T>
static void DumpResourceSection (const BYTE *pImageBase,
                                 const T    *pNTHeader)
{
  const BYTE *pResDir;
  BOOL        bIs64Bit     = (pNTHeader->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC);
  DWORD       resourcesRVA = GetImgDirEntryRVA (pNTHeader, IMAGE_DIRECTORY_ENTRY_RESOURCE);

  if (!resourcesRVA)
    return;

  pResDir = (const BYTE*) GetPtrFromRVA (resourcesRVA, pNTHeader, pImageBase);
  if (!pResDir)
    return;

  printf ("Resources (RVA: %lX)\n", resourcesRVA);

  DumpResourceDirectory ((const IMAGE_RESOURCE_DIRECTORY*)pResDir, pResDir, 0, 0);

  printf ("\n");

  if (!fShowResources)
     return;

  if (g_cStrResEntries)
  {
    printf ("String Table\n");

    if (bIs64Bit)
         DumpStringTable (pImageBase, (const IMAGE_NT_HEADERS64*)pNTHeader, pResDir, g_pStrResEntries, g_cStrResEntries);
    else DumpStringTable (pImageBase, (const IMAGE_NT_HEADERS32*)pNTHeader, pResDir, g_pStrResEntries, g_cStrResEntries);

    printf ("\n");
  }

  if (g_cDlgResEntries)
  {
    printf ("Dialogs\n");

    if (bIs64Bit)
         DumpDialogs (pImageBase, (const IMAGE_NT_HEADERS64*) pNTHeader, pResDir, g_pDlgResEntries, g_cDlgResEntries);
    else DumpDialogs (pImageBase, (const IMAGE_NT_HEADERS32*) pNTHeader, pResDir, g_pDlgResEntries, g_cDlgResEntries);

    printf ("\n");
  }

  printf ("\n");
}

void DumpResources (const BYTE *pImageBase, const IMAGE_NT_HEADERS *pNTHeader)
{
  if (pNTHeader->OptionalHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC)
       DumpResourceSection (pImageBase, (const IMAGE_NT_HEADERS64*)pNTHeader);
  else DumpResourceSection (pImageBase, (const IMAGE_NT_HEADERS32*)pNTHeader);
}
